/*************************************************************************/ /*
 VSPM

 Copyright (C) 2013-2014 Renesas Electronics Corporation

 License        Dual MIT/GPLv2

 The contents of this file are subject to the MIT license as set out below.

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 Alternatively, the contents of this file may be used under the terms of
 the GNU General Public License Version 2 ("GPL") in which case the provisions
 of GPL are applicable instead of those above.

 If you wish to allow use of your version of this file only under the terms of
 GPL, and not to allow others to use your version of this file under the terms
 of the MIT license, indicate your decision by deleting the provisions above
 and replace them with the notice and other provisions required by GPL as set
 out in the file called "GPL-COPYING" included in this distribution. If you do
 not delete the provisions above, a recipient may use your version of this file
 under the terms of either the MIT license or GPL.

 This License is also included in this distribution in the file called
 "MIT-COPYING".

 EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
 PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
 BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
 COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


 GPLv2:
 If you wish to use this file under the terms of GPL, following terms are
 effective.

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; version 2 of the License.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/ /*************************************************************************/

#include <linux/version.h>
#include <asm/uaccess.h>
#include <linux/module.h>
#include <linux/cdev.h>
#include <linux/fs.h>
#include <linux/platform_device.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/dma-mapping.h>
#include <linux/list.h>
#include <linux/proc_fs.h>
#include <linux/sched.h>
#include <linux/completion.h>
#include <linux/pm_runtime.h>
#include <linux/of_address.h>
#include <linux/miscdevice.h>

#include "vspm_public.h"
#include "vspm_private.h"
#include "vspm_main.h"
#include "vspm_log.h"


long debug_flg = 0xff;

struct vspm_drvdata *p_vspm_drvdata;

/**
 * VSPM_lib_DriverInitialize - Initialize the VSPM driver
 * @handle: destination address of the handle
 * Description: Initialize the driver of each IP (VSPS, 2DDAMC).
 * Returns: On success R_VSPM_OK is returned. On error, R_VSPM_NG is returned.
 */
long VSPM_lib_DriverInitialize(unsigned long *handle)
{
	int ercd;
	struct vspm_privdata *priv = 0;
	struct vspm_drvdata *pdrv = p_vspm_drvdata;

	down(&pdrv->init_sem);
	DPRINT("called\n");

	/* check parameter */
	if (handle == NULL)
		goto err_exit1;

	/* allocate the private data area of the vspm device file */
	priv = kzalloc(sizeof(*priv), GFP_KERNEL);
	if (!priv) {
		APRINT("could not allocate the private data area of the ");
		APRINT("vspm device file\n");
		goto err_exit1;
	}

	spin_lock_init(&priv->lock);

	init_completion(&priv->comp);
	init_completion(&priv->cb_start_comp);

	INIT_LIST_HEAD(&priv->req_head.list);
	INIT_LIST_HEAD(&priv->rsp_head.list);

	priv->pdrv = pdrv;

	if (atomic_add_return(1, &pdrv->counter) == 1) {
		DPRINT("first time open\n");
		/* first time open */
		ercd = vspm_init(priv);
		if (ercd) {
			APRINT("failed to vspm_init %d\n", ercd);
			goto err_exit2;
		}
	} else {
		DPRINT("already opened\n");
	}

	*handle = (unsigned long)priv;

	DPRINT("done handle=%08x\n", (unsigned int)*handle);
	up(&pdrv->init_sem);
	return R_VSPM_OK;

err_exit2:
	atomic_dec(&pdrv->counter);
	kfree(priv);

err_exit1:
	up(&pdrv->init_sem);
	return R_VSPM_NG;
}
EXPORT_SYMBOL(VSPM_lib_DriverInitialize);

/*
 * vspm_open - Open the VSPM driver
 * Returns: On success 0 is returned. On error, a nonzero error number is
 * returned.
 */
static int vspm_open(struct inode *inode, struct file *filp)
{
	long ercd;
	long handle;

	DPRINT("called\n");

	ercd = VSPM_lib_DriverInitialize(&handle);
	if (ercd != R_VSPM_OK) {
		APRINT("failed to VSPM_lib_DriverInitialize %d\n", (int)ercd);
		return -EFAULT;
	}

	filp->private_data = (void *)handle;

	DPRINT("done\n");
	return 0;
}

/**
 * VSPM_lib_DriverQuit - Exit the VSPM driver
 * @handle: handle
 * Description: Exit the driver of each IP (VSPS, 2DDAMC).
 * Returns: On success R_VSPM_OK is returned. On error, R_VSPM_NG is returned.
 */
long VSPM_lib_DriverQuit(unsigned long handle)
{
	int ercd;
	struct vspm_privdata *priv = (struct vspm_privdata *)handle;
	struct vspm_drvdata *pdrv = p_vspm_drvdata;
	int cnt;

	down(&pdrv->init_sem);
	DPRINT("called\n");

	/* check parameter */
	if (priv == NULL)
		goto err_exit1;

	if (priv->pdrv != pdrv)
		goto err_exit1;

	cnt = atomic_sub_return(1, &pdrv->counter);
	if (cnt == 0) {
		DPRINT("last close\n");
		ercd = vspm_quit(priv);
		if (ercd) {
			APRINT("failed to vspm_quit %d\n", ercd);
			goto err_exit2;
		}
	} else if (cnt > 0) {
		DPRINT("still open\n");
		ercd = vspm_cancel(priv);
		if (ercd) {
			APRINT("failed to vspm_cancel %d\n", ercd);
			goto err_exit2;
		}
	} else {
		APRINT("already closed\n");
		goto err_exit2;
	}

	up(&pdrv->init_sem);
	priv->pdrv = NULL;
	kfree(priv);

	DPRINT("done\n");
	return R_VSPM_OK;

err_exit2:
	atomic_inc(&pdrv->counter);

err_exit1:
	up(&pdrv->init_sem);
	return R_VSPM_NG;
}
EXPORT_SYMBOL(VSPM_lib_DriverQuit);

/*
 * vspm_release - Release the VSPM driver
 * Returns: On success 0 is returned. On error, a nonzero error number is
 * returned.
 */
static int vspm_release(struct inode *inode, struct file *filp)
{
	long ercd;
	long handle = (long)filp->private_data;

	DPRINT("called\n");

	ercd = VSPM_lib_DriverQuit(handle);
	if (ercd != R_VSPM_OK) {
		APRINT("failed to VSPM_lib_DriverQuit %d\n", (int)ercd);
		return -EFAULT;
	}

	filp->private_data = NULL;

	DPRINT("done\n");
	return 0;
}


static const struct file_operations vspm_fops = {
	.owner			= THIS_MODULE,
	.unlocked_ioctl	= vspm_ioctl,
	.open			= vspm_open,
	.release		= vspm_release,
};

static struct miscdevice vspm_misc = {
	.minor = MISC_DYNAMIC_MINOR,
	.name  = DEVNAME,
	.fops  = &vspm_fops
};

#define VSPM_TDDMAC 	VSPM_IP_MAX

static const struct of_device_id vspm_ip_matches[] = {
	{ .compatible = "renesas,vspm-vsps", .data = VSPM_IP_VSPS },
	{ .compatible = "renesas,vspm-vspr", .data = VSPM_IP_VSPR },
	{ .compatible = "renesas,vspm-vspd0", .data = VSPM_IP_VSPD0 },
	{ .compatible = "renesas,vspm-vspd1", .data = VSPM_IP_VSPD1 },
	{ .compatible = "renesas,vspm-tddmac", .data = VSPM_TDDMAC },
	{ /* sentinel */ }
};




/*
 * vspm_probe - Probe the VSPM driver
 * Returns: On success 0 is returned. On error, a nonzero error number is
 * returned.
 */
static int vspm_probe(struct platform_device *pdev)
{
	struct vspm_drvdata *pdrv = p_vspm_drvdata;
	const struct of_device_id *match;
	struct platform_device **p_pdev;
	struct device_node *np;
	
	int ch;

	DPRINT("called(id=%d name=%s)\n", pdev->id, pdev->name);

	match = of_match_node(vspm_ip_matches, pdev->dev.of_node);
	if (!match)
		return -ENODEV;

	/* save platform_device pointer */
	ch = (int)match->data;
	if (ch == VSPM_TDDMAC){
		pdrv->tddmac_pdev = pdev;
		p_pdev = &pdrv->tddmac_pdev;
	}else{
		pdrv->vsp_pdev[ch] = pdev;
		p_pdev = &pdrv->vsp_pdev[ch];
	}

	DPRINT("called(ch=%d)\n", ch);

	/* save drvdata */
	platform_set_drvdata(pdev, p_pdev);

	pm_suspend_ignore_children(&pdev->dev, true);
	pm_runtime_enable(&pdev->dev);
	pm_runtime_get_sync(&pdev->dev);

	DPRINT("done(id=%d)\n", pdev->id);
	return 0;
}

/*
 * vspm_remove - Remove the VSPM driver
 * Returns: On success 0 is returned. On error, a nonzero error number is
 * returned.
 */
static int vspm_remove(struct platform_device *pdev)
{
	struct vspm_drvdata **pdrv = platform_get_drvdata(pdev);

	DPRINT("called(id=%d)\n", pdev->id);

	pm_runtime_disable(&pdev->dev);
	*pdrv = NULL;
	platform_set_drvdata(pdev, NULL);

	DPRINT("done(id=%d)\n", pdev->id);
	return 0;
}

static int vspm_runtime_nop(struct device *dev)
{
	/* Runtime PM callback shared between ->runtime_suspend()
	 * and ->runtime_resume(). Simply returns success.
	 *
	 * This driver re-initializes all registers after
	 * pm_runtime_get_sync() anyway so there is no need
	 * to save and restore registers here.
	 */
	return 0;
}

static const struct dev_pm_ops vspm_pm_ops = {
	.runtime_suspend = vspm_runtime_nop,
	.runtime_resume = vspm_runtime_nop,
	.suspend = vspm_runtime_nop,
	.resume = vspm_runtime_nop,
};


static struct platform_driver vspm_driver = {
	.driver		= {
		.name		= DRVNAME,
		.owner		= THIS_MODULE,
		.pm		= &vspm_pm_ops,
		.of_match_table = vspm_ip_matches,
	},
	.probe		= vspm_probe,
	.remove		= vspm_remove,
};


/* vspm device release function */
static void vspm_dev_release(struct device *dev)
{
	DPRINT("do not do anything\n");
	return;
}


static int __init vspm_module_init(void)
{
	struct vspm_drvdata *pdrv = 0;
	int ercd = 0;


	DPRINT("called\n");

	/* allocate vspm driver data area */
	pdrv = kzalloc(sizeof(*pdrv), GFP_KERNEL);
	if (!pdrv) {
		APRINT("could not allocate vspm driver data area\n");
		return -ENOMEM;
	}
	p_vspm_drvdata = pdrv;

	/* register a driver for platform-level devices */
	ercd = platform_driver_register(&vspm_driver);
	if (ercd) {
		kfree(pdrv);
		p_vspm_drvdata = NULL;
		return ercd;
	}

	/* initialize open counter */
	atomic_set(&pdrv->counter, 0);

	/* initialize semaphore */
	sema_init(&pdrv->init_sem, 1);/* unlock */

	ercd = misc_register(&vspm_misc);
	if (ercd) {
		platform_driver_unregister(&vspm_driver);
		kfree(pdrv);
		p_vspm_drvdata = NULL;
		APRINT("misc_register Fail\n");
	}

	DPRINT("done\n");

	return 0;
}


static void __exit vspm_module_exit(void)
{
	struct vspm_drvdata *pdrv = p_vspm_drvdata;

	DPRINT("called\n");

	misc_deregister(&vspm_misc);
	platform_driver_unregister(&vspm_driver);

	kfree(pdrv);
	p_vspm_drvdata = NULL;

	DPRINT("done\n");
}

module_init(vspm_module_init);
module_exit(vspm_module_exit);


MODULE_AUTHOR("Renesas Electronics Corporation");
MODULE_LICENSE("Dual MIT/GPL");

